"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const pump = require("pump");
const url_1 = require("url");
const tough_cookie_1 = require("tough-cookie");
const throwback_1 = require("throwback");
const http_1 = require("http");
const https_1 = require("https");
const net_1 = require("net");
const tls_1 = require("tls");
const http2_1 = require("http2");
const stream_1 = require("stream");
const zlib_1 = require("zlib");
const servie_1 = require("servie");
const node_1 = require("servie/dist/body/node");
const error_1 = require("../error");
const common_1 = require("../common");
/**
 * HTTP responses implement a node.js body.
 */
class HttpResponse extends servie_1.Response {
    constructor(options) {
        super(options);
        this.url = options.url;
        this.body = options.body;
        this.httpVersion = options.httpVersion;
    }
}
exports.HttpResponse = HttpResponse;
class Http2Response extends HttpResponse {
}
exports.Http2Response = Http2Response;
/**
 * Read cookies from the cookie jar.
 */
function getCookies(options) {
    return function (req, next) {
        const prevCookies = req.headers.getAll('Cookie').join('; ');
        return new Promise((resolve, reject) => {
            options.jar.getCookieString(req.url, (err, cookies) => {
                if (err)
                    return reject(err);
                if (cookies) {
                    req.headers.set('Cookie', prevCookies ? `${prevCookies}; ${cookies}` : cookies);
                }
                return resolve(next());
            });
        });
    };
}
exports.getCookies = getCookies;
/**
 * Save response cookies into the cookie jar.
 */
function saveCookies(options) {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield next();
            const cookies = res.headers.getAll('set-cookie');
            if (cookies.length) {
                yield Promise.all(cookies.map(function (cookie) {
                    return new Promise(function (resolve, reject) {
                        options.jar.setCookie(cookie, req.url, { ignoreError: true }, (err) => err ? reject(err) : resolve());
                    });
                }));
            }
            return res;
        });
    };
}
exports.saveCookies = saveCookies;
/**
 * Automatically support decoding zipped responses.
 */
function autoUnzip() {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!req.headers.has('Accept-Encoding')) {
                req.headers.set('Accept-Encoding', 'gzip,deflate');
            }
            const res = yield next();
            const enc = res.headers.get('Content-Encoding');
            // Unzip body automatically when response is encoded.
            if (enc === 'deflate' || enc === 'gzip') {
                res.body = node_1.createBody(pump(res.body.stream(), zlib_1.createUnzip()), { headers: [] });
            }
            return res;
        });
    };
}
exports.autoUnzip = autoUnzip;
/**
 * Set default user-agent in node.js.
 */
function normalizeUserAgent(options = {}) {
    const userAgent = options.userAgent || 'Popsicle (https://github.com/serviejs/popsicle)';
    return function (req, next) {
        if (!req.headers.has('User-Agent'))
            req.headers.set('User-Agent', userAgent);
        return next();
    };
}
exports.normalizeUserAgent = normalizeUserAgent;
/**
 * Redirection types to handle.
 */
var REDIRECT_TYPE;
(function (REDIRECT_TYPE) {
    REDIRECT_TYPE[REDIRECT_TYPE["FOLLOW_WITH_GET"] = 0] = "FOLLOW_WITH_GET";
    REDIRECT_TYPE[REDIRECT_TYPE["FOLLOW_WITH_CONFIRMATION"] = 1] = "FOLLOW_WITH_CONFIRMATION";
})(REDIRECT_TYPE || (REDIRECT_TYPE = {}));
/**
 * Possible redirection status codes.
 */
const REDIRECT_STATUS = {
    '301': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '302': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '303': REDIRECT_TYPE.FOLLOW_WITH_GET,
    '307': REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION,
    '308': REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION
};
/**
 * Middleware function for following HTTP redirects.
 */
function followRedirects(process, options = {}) {
    return function (initialRequest, done) {
        return __awaiter(this, void 0, void 0, function* () {
            let req = initialRequest.clone();
            let redirectCount = 0;
            const maxRedirects = typeof options.maxRedirects === 'number' ? options.maxRedirects : 5;
            const confirmRedirect = options.confirmRedirect || (() => false);
            while (redirectCount++ < maxRedirects) {
                const res = yield process(req, done);
                const redirect = REDIRECT_STATUS[res.statusCode];
                if (redirect === undefined || !res.headers.has('Location'))
                    return res;
                const newUrl = url_1.resolve(req.url, res.headers.get('Location')); // tslint:disable-line
                // Ignore the result of the response on redirect.
                req.abort();
                req.events.emit('redirect', newUrl);
                if (redirect === REDIRECT_TYPE.FOLLOW_WITH_GET) {
                    req = initialRequest.clone();
                    req.headers.set('Content-Length', '0');
                    req.url = newUrl;
                    req.method = req.method.toUpperCase() === 'HEAD' ? 'HEAD' : 'GET';
                    req.body = node_1.createBody(undefined);
                    req.trailer = Promise.resolve(servie_1.createHeaders());
                    continue;
                }
                if (redirect === REDIRECT_TYPE.FOLLOW_WITH_CONFIRMATION) {
                    const method = req.method.toUpperCase();
                    // Following HTTP spec by automatically redirecting with GET/HEAD.
                    if (method === 'GET' || method === 'HEAD') {
                        req = initialRequest.clone();
                        req.url = newUrl;
                        continue;
                    }
                    // Allow the user to confirm redirect according to HTTP spec.
                    if (confirmRedirect(req, res)) {
                        req = initialRequest.clone();
                        req.url = newUrl;
                        continue;
                    }
                }
                return res;
            }
            throw new error_1.PopsicleError(`Maximum redirects exceeded: ${maxRedirects}`, 'EMAXREDIRECTS', req);
        });
    };
}
exports.followRedirects = followRedirects;
/**
 * Track HTTP connections for reuse.
 */
class ConnectionManager {
    constructor() {
        this.connections = new Map();
    }
    get(key) {
        return this.connections.get(key);
    }
    set(key, connection) {
        if (this.connections.has(key))
            throw new TypeError('Connection exists for key');
        this.connections.set(key, connection);
        return connection;
    }
    delete(key, connection) {
        const existing = this.connections.get(key);
        if (existing !== connection)
            throw new TypeError('Connection for key does not match');
        this.connections.delete(key);
        return connection;
    }
}
exports.ConnectionManager = ConnectionManager;
/**
 * Manage HTTP connection reuse.
 */
class ConcurrencyConnectionManager extends ConnectionManager {
    constructor(options = {}) {
        super();
        this.options = options;
        this.maxConnections = Infinity;
        this.maxFreeConnections = 256;
        if (options.maxConnections)
            this.maxConnections = options.maxConnections;
        if (options.maxFreeConnections)
            this.maxFreeConnections = options.maxFreeConnections;
    }
    /**
     * Create a new connection.
     */
    ready(key, onReady) {
        const pool = this.get(key) || this.set(key, Object.create(null));
        // Reuse free connections first.
        if (pool.free)
            return onReady(this.getFreeConnection(key));
        // If no other connections exist, `onReady` immediately.
        if (!pool.used)
            return onReady();
        // Add to "pending" queue.
        if (pool.used.size >= this.maxConnections) {
            if (!pool.pend)
                pool.pend = [];
            pool.pend.push(onReady);
            return;
        }
        return onReady();
    }
    getUsedConnection(key) {
        const pool = this.get(key);
        if (pool && pool.used)
            return pool.used.values().next().value;
    }
    getFreeConnection(key) {
        const pool = this.get(key);
        if (pool && pool.free)
            return pool.free.values().next().value;
    }
    use(key, connection) {
        const pool = this.get(key) || this.set(key, Object.create(null));
        if (pool.free)
            pool.free.delete(connection);
        if (!pool.used)
            pool.used = new Set();
        pool.used.add(connection);
    }
    freed(key, connection, discard) {
        const pool = this.get(key) || this.set(key, Object.create(null));
        // Remove from any possible "used".
        if (pool.used)
            pool.used.delete(connection);
        // Immediately send for connection.
        if (pool.pend) {
            const onReady = pool.pend.shift();
            onReady(connection);
            if (!pool.pend.length)
                delete pool.pend;
        }
        // Add to "free" connections pool.
        if (!pool.free)
            pool.free = new Set();
        if (pool.free.size >= this.maxFreeConnections)
            return discard();
        pool.free.add(connection);
    }
    remove(key, connection) {
        const pool = this.get(key);
        if (!pool)
            return;
        if (pool.used && pool.used.has(connection)) {
            pool.used.delete(connection);
            if (!pool.used.size)
                delete pool.used;
        }
        if (pool.free && pool.free.has(connection)) {
            pool.free.delete(connection);
            if (!pool.free.size)
                delete pool.free;
        }
        if (!pool.free && !pool.used && !pool.pend)
            this.delete(key, pool);
    }
}
exports.ConcurrencyConnectionManager = ConcurrencyConnectionManager;
/**
 * Configure HTTP version negotiation.
 */
var NegotiateHttpVersion;
(function (NegotiateHttpVersion) {
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP1_ONLY"] = 0] = "HTTP1_ONLY";
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP2_FOR_HTTPS"] = 1] = "HTTP2_FOR_HTTPS";
    NegotiateHttpVersion[NegotiateHttpVersion["HTTP2_ONLY"] = 2] = "HTTP2_ONLY";
})(NegotiateHttpVersion = exports.NegotiateHttpVersion || (exports.NegotiateHttpVersion = {}));
/**
 * Write Servie body object to node.js stream.
 */
function sendBody(body, stream, onError) {
    // Pipe the body to the stream.
    if (body instanceof node_1.Body) {
        if (body.buffered) {
            body.buffer().then(x => stream.end(x), onError);
        }
        else {
            pump(body.stream(), stream);
        }
    }
    else {
        body.arrayBuffer().then(x => stream.end(Buffer.from(x)), onError);
    }
}
// Global connection caches.
const globalNetConnections = new ConcurrencyConnectionManager();
const globalTlsConnections = new ConcurrencyConnectionManager();
const globalHttp2Connections = new ConnectionManager();
/**
 * Execute HTTP request.
 */
function execHttp1(req, protocol, host, port, keepAlive, socket) {
    return new Promise((resolve, reject) => {
        const { url, body, Url } = req;
        const encrypted = Url.protocol === 'https:';
        const request = encrypted ? https_1.request : http_1.request;
        const arg = {
            protocol,
            host,
            port,
            method: req.method,
            path: Url.path,
            headers: req.headers.asObject(false),
            auth: Url.auth,
            createConnection: () => socket
        };
        const rawRequest = request(arg);
        const requestStream = new stream_1.PassThrough();
        // Reuse HTTP connections where possible.
        if (keepAlive > 0) {
            rawRequest.shouldKeepAlive = true;
            rawRequest.setHeader('Connection', 'keep-alive');
        }
        // Trigger unavailable error when node.js errors before response.
        function onError(err) {
            return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
        }
        // Track the node.js response.
        function onResponse(rawResponse) {
            const { statusCode, statusMessage, httpVersion } = rawResponse;
            const headers = servie_1.createHeaders(rawResponse.rawHeaders);
            const body = node_1.createBody(pump(rawResponse, new stream_1.PassThrough()), { headers: {} });
            // Trailers are populated on "end".
            const trailer = new Promise(resolve => {
                rawResponse.once('end', () => resolve(servie_1.createHeaders(rawResponse.rawTrailers)));
            });
            // Replace request error listener behaviour.
            rawRequest.removeListener('error', onError);
            rawRequest.on('error', err => req.events.emit('error', err));
            const { address: localAddress, port: localPort } = rawRequest.connection.address();
            const { address: remoteAddress, port: remotePort } = rawResponse.connection.address();
            const res = new HttpResponse({ statusCode, statusMessage, headers, trailer, body, url, httpVersion });
            // Update request connection.
            req.connection = { localAddress, localPort, remoteAddress, remotePort, encrypted };
            // https://github.com/serviejs/servie#implementers
            res.started = true;
            req.events.emit('response', res);
            // Track response progress.
            rawResponse.on('data', (chunk) => res.bytesTransferred += chunk.length);
            rawResponse.once('end', () => {
                req.closed = true;
                res.finished = true;
            });
            return resolve(res);
        }
        rawRequest.once('error', onError);
        rawRequest.once('response', onResponse);
        // https://github.com/serviejs/servie#implementers
        req.started = true;
        req.events.once('abort', () => {
            socket.emit('agentRemove'); // `abort` destroys the connection with no event.
            rawRequest.abort();
        });
        // Track request upload progress.
        requestStream.on('data', (chunk) => req.bytesTransferred += chunk.length);
        pump(requestStream, rawRequest, () => req.finished = true);
        return sendBody(body, requestStream, reject);
    });
}
/**
 * Execute a HTTP2 connection.
 */
function execHttp2(req, protocol, host, port, client) {
    return new Promise((resolve, reject) => {
        // HTTP2 formatted headers.
        const headers = Object.assign(req.headers.asObject(false), {
            [http2_1.constants.HTTP2_HEADER_METHOD]: req.method,
            [http2_1.constants.HTTP2_HEADER_PATH]: req.Url.path
        });
        const http2Stream = client.request(headers, { endStream: false });
        const requestStream = new stream_1.PassThrough();
        ref(client.socket); // Request ref tracking.
        // Track when stream finishes.
        function onClose() {
            req.closed = true;
            unref(client.socket);
        }
        // Trigger unavailable error when node.js errors before response.
        function onError(err) {
            return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
        }
        function onResponse(headers) {
            const encrypted = client.socket.encrypted === true;
            const { localAddress, localPort, remoteAddress, remotePort } = client.socket;
            // Replace request error listener behaviour with proxy.
            http2Stream.removeListener('error', onError);
            http2Stream.on('error', err => req.events.emit('error', err));
            const res = new Http2Response({
                statusCode: Number(headers[http2_1.constants.HTTP2_HEADER_STATUS]),
                url: req.url,
                httpVersion: '2.0',
                headers: servie_1.createHeaders(headers),
                body: node_1.createBody(pump(http2Stream, new stream_1.PassThrough()), { headers: {} })
            });
            // https://github.com/serviejs/servie#implementers
            res.started = true;
            req.connection = { localAddress, localPort, remoteAddress, remotePort, encrypted };
            // Track response progress.
            http2Stream.on('data', (chunk) => res.bytesTransferred += chunk.length);
            http2Stream.once('end', () => res.finished = true);
            return resolve(res);
        }
        http2Stream.once('error', onError);
        http2Stream.once('close', onClose);
        http2Stream.once('response', onResponse);
        // https://github.com/serviejs/servie#implementers
        req.started = true;
        req.events.once('abort', () => http2Stream.destroy());
        // Track request upload progress.
        requestStream.on('data', (chunk) => req.bytesTransferred += chunk.length);
        pump(requestStream, http2Stream, () => req.finished = true);
        return sendBody(req.body, requestStream, reject);
    });
}
/**
 * Forward request over `http`.
 */
function forward(options) {
    const { keepAlive = 5000, // Default to keeping a connection open briefly.
    negotiateHttpVersion = NegotiateHttpVersion.HTTP2_FOR_HTTPS } = options;
    // TODO: Allow configuration in options.
    const tlsConnections = globalTlsConnections;
    const netConnections = globalNetConnections;
    const http2Connections = globalHttp2Connections;
    return function (req) {
        const { hostname, protocol } = req.Url;
        const host = hostname || 'localhost';
        if (protocol === 'http:') {
            const port = Number(req.Url.port) || 80;
            const connectionKey = `${host}:${port}:${negotiateHttpVersion}`;
            // Use existing HTTP2 session in HTTP2 mode.
            if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                const existingSession = http2Connections.get(connectionKey);
                if (existingSession)
                    return execHttp2(req, protocol, host, port, existingSession);
            }
            return new Promise((resolve) => {
                return netConnections.ready(connectionKey, (freeSocket) => {
                    const socketOptions = { host, port };
                    const socket = freeSocket || setupSocket(connectionKey, keepAlive, netConnections, net_1.connect(socketOptions));
                    socket.ref();
                    netConnections.use(connectionKey, socket);
                    if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                        const authority = `${protocol}//${host}:${port}`;
                        const client = manageHttp2(authority, connectionKey, keepAlive, http2Connections, socket);
                        return resolve(execHttp2(req, protocol, host, port, client));
                    }
                    return resolve(execHttp1(req, protocol, host, port, keepAlive, socket));
                });
            });
        }
        // Optionally negotiate HTTP2 connection.
        if (protocol === 'https:') {
            const { ca, cert, key, secureProtocol, secureContext } = options;
            const port = Number(req.Url.port) || 443;
            const servername = options.servername || calculateServerName(host, req.headers.get('host'));
            const rejectUnauthorized = options.rejectUnauthorized !== false;
            const connectionKey = `${host}:${port}:${negotiateHttpVersion}:${servername}:${rejectUnauthorized}:${ca || ''}:${cert || ''}:${key || ''}:${secureProtocol || ''}`;
            // Use an existing TLS session to speed up handshake.
            const existingSocket = tlsConnections.getFreeConnection(connectionKey) || tlsConnections.getUsedConnection(connectionKey);
            const session = existingSocket ? existingSocket.getSession() : undefined;
            const socketOptions = {
                host, port, servername, rejectUnauthorized, ca, cert, key,
                session, secureProtocol, secureContext
            };
            // Use any existing HTTP2 session.
            if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY ||
                negotiateHttpVersion === NegotiateHttpVersion.HTTP2_FOR_HTTPS) {
                const existingSession = http2Connections.get(connectionKey);
                if (existingSession)
                    return execHttp2(req, protocol, host, port, existingSession);
            }
            return new Promise((resolve, reject) => {
                // Set up ALPN protocols for connection negotiation.
                if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                    socketOptions.ALPNProtocols = ['h2'];
                }
                else if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_FOR_HTTPS) {
                    socketOptions.ALPNProtocols = ['h2', 'http/1.1'];
                }
                return tlsConnections.ready(connectionKey, (freeSocket) => {
                    const socket = freeSocket || setupSocket(connectionKey, keepAlive, tlsConnections, tls_1.connect(socketOptions));
                    socket.ref();
                    tlsConnections.use(connectionKey, socket);
                    if (negotiateHttpVersion === NegotiateHttpVersion.HTTP1_ONLY) {
                        return resolve(execHttp1(req, protocol, host, port, keepAlive, socket));
                    }
                    if (negotiateHttpVersion === NegotiateHttpVersion.HTTP2_ONLY) {
                        const client = manageHttp2(`${protocol}//${host}:${port}`, connectionKey, keepAlive, http2Connections, socket);
                        return resolve(execHttp2(req, protocol, host, port, client));
                    }
                    socket.once('secureConnect', () => {
                        const alpnProtocol = socket.alpnProtocol;
                        // Successfully negotiated HTTP2 connection.
                        if (alpnProtocol === 'h2') {
                            const existingClient = http2Connections.get(connectionKey);
                            if (existingClient) {
                                socket.destroy(); // Destroy socket in case of TLS connection race.
                                return resolve(execHttp2(req, protocol, host, port, existingClient));
                            }
                            const client = manageHttp2(`${protocol}//${host}:${port}`, connectionKey, keepAlive, http2Connections, socket);
                            return resolve(execHttp2(req, protocol, host, port, client));
                        }
                        if (alpnProtocol === 'http/1.1') {
                            return resolve(execHttp1(req, protocol, host, port, keepAlive, socket));
                        }
                        return reject(new error_1.PopsicleError('No ALPN protocol negotiated', 'EALPNPROTOCOL', req));
                    });
                    socket.once('error', (err) => {
                        return reject(new error_1.PopsicleError(`Unable to connect to ${host}:${port}`, 'EUNAVAILABLE', req, err));
                    });
                });
            });
        }
        return Promise.reject(new error_1.PopsicleError(`Unsupported URL protocol: ${req.Url.protocol}`, 'EPROTOCOL', req));
    };
}
exports.forward = forward;
/**
 * Setup the socket with the connection manager.
 *
 * Ref: https://github.com/nodejs/node/blob/531b4bedcac14044f09129ffb65dab71cc2707d9/lib/_http_agent.js#L254
 */
function setupSocket(key, keepAlive, manager, socket) {
    const onFree = () => {
        if (keepAlive > 0) {
            socket.setKeepAlive(true, keepAlive);
            socket.unref();
        }
        manager.freed(key, socket, () => socket.destroy());
    };
    const onClose = () => manager.remove(key, socket);
    const onRemove = () => {
        socket.removeListener('free', onFree);
        socket.removeListener('close', onClose);
        manager.remove(key, socket);
    };
    socket.on('free', onFree);
    socket.once('close', onClose);
    socket.once('agentRemove', onRemove);
    return socket;
}
/**
 * Set up a HTTP2 working session.
 */
function manageHttp2(authority, key, keepAlive, manager, socket) {
    // TODO: Fix node.js types.
    const connectOptions = { createConnection: () => socket };
    const client = http2_1.connect(authority, connectOptions);
    manager.set(key, client);
    client.once('close', () => manager.delete(key, client));
    client.setTimeout(keepAlive, () => client.close());
    return client;
}
/**
 * Track socket usage.
 */
const SOCKET_REFS = new WeakMap();
/**
 * Track socket refs.
 */
function ref(socket) {
    const count = SOCKET_REFS.get(socket) || 0;
    if (count === 0)
        socket.ref();
    SOCKET_REFS.set(socket, count + 1);
}
/**
 * Track socket unrefs and globally unref.
 */
function unref(socket) {
    const count = SOCKET_REFS.get(socket);
    if (!count)
        return;
    if (count === 1) {
        socket.unref();
        SOCKET_REFS.delete(socket);
        return;
    }
    SOCKET_REFS.set(socket, count - 1);
}
/**
 * Create a request transport using node.js `http` libraries.
 */
function transport(options = {}) {
    const fns = [];
    const { jar = new tough_cookie_1.CookieJar(), unzip = true, follow = true } = options;
    // Built-in behaviours.
    fns.push(common_1.normalizeRequest(options));
    fns.push(normalizeUserAgent(options));
    if (unzip)
        fns.push(autoUnzip());
    if (jar)
        fns.push(getCookies({ jar }), saveCookies({ jar }));
    const done = forward(options);
    const middleware = follow ? followRedirects(throwback_1.compose(fns), options) : throwback_1.compose(fns);
    return (req) => middleware(req, done);
}
exports.transport = transport;
/**
 * Ref: https://github.com/nodejs/node/blob/5823938d156f4eb6dc718746afbf58f1150f70fb/lib/_http_agent.js#L231
 */
function calculateServerName(host, hostHeader) {
    if (!hostHeader)
        return host;
    if (hostHeader.charAt(0) === '[') {
        const index = hostHeader.indexOf(']');
        if (index === -1)
            return hostHeader;
        return hostHeader.substr(1, index - 1);
    }
    return hostHeader.split(':', 1)[0];
}
//# sourceMappingURL=http.js.map