import { Request, Response, ResponseOptions } from 'servie';
import { Body } from 'servie/dist/body/browser';
/**
 * Extend response with URL.
 */
export interface XhrResponseOptions extends ResponseOptions {
    url: string;
    body: Body;
}
/**
 * XHR responses can indicate a response URL.
 */
export declare class XhrResponse extends Response implements XhrResponseOptions {
    url: string;
    body: Body;
    constructor(options: XhrResponseOptions);
}
/**
 * Valid XHR configuration.
 */
export interface ForwardOptions {
    type?: XMLHttpRequestResponseType;
    withCredentials?: boolean;
    overrideMimeType?: string;
}
/**
 * Forward request over `XMLHttpRequest`.
 */
export declare function forward(options: ForwardOptions): (req: Request) => Promise<XhrResponse>;
/**
 * Combined transport options.
 */
export interface TransportOptions extends ForwardOptions {
}
/**
 * Create a request transport using `XMLHttpRequest`.
 */
export declare function transport(options?: TransportOptions): (req: Request) => Response | Promise<Response>;
